@@ -5,17 +5,17 @@
 # Compare files created by /usr/src/etc/Makefile (or the directory
 # the user specifies) with the currently installed copies.
 
-# Copyright 1998-2002 Douglas Barton
+# Copyright 1998-2003 Douglas Barton
 # DougB@FreeBSD.org
 
-# $FreeBSD: src/usr.sbin/mergemaster/mergemaster.sh,v 1.26 2002/02/12 09:54:56 dougb Exp $
+# $FreeBSD: mergemaster.sh,v 1.46 2003/05/03 06:35:19 dougb Exp $
 
 PATH=/bin:/usr/bin:/usr/sbin
 
 display_usage () {
   VERSION_NUMBER=`grep "[$]FreeBSD:" $0 | cut -d ' ' -f 4`
   echo "mergemaster version ${VERSION_NUMBER}"
-  echo 'Usage: mergemaster [-scrvahi] [-m /path]'
+  echo 'Usage: mergemaster [-scrvahipCP] [-m /path]'
   echo '         [-t /path] [-d] [-u N] [-w N] [-D /path]'
   echo "Options:"
   echo "  -s  Strict comparison (diff every pair of files)"
@@ -25,6 +25,9 @@ display_usage () {
   echo "  -a  Leave all files that differ to merge by hand"
   echo "  -h  Display more complete help"
   echo '  -i  Automatically install files that do not exist in destination directory'
+  echo '  -p  Pre-buildworld mode, only compares crucial files'
+  echo '  -C  Compare local rc.conf variables to the defaults'
+  echo '  -P  Preserve files that are overwritten'
   echo "  -m /path/directory  Specify location of source to do the make in"
   echo "  -t /path/directory  Specify temp root directory"
   echo "  -d  Add date and time to directory name (e.g., /var/tmp/temproot.`date +%m%d.%H.%M`)"
@@ -113,16 +116,16 @@ diff_loop () {
 	echo '   ======================================================================   '
 	echo ''
         (
-          echo ''
           echo "  *** Displaying differences between ${COMPFILE} and installed version:"
           echo ''
-          diff "${DIFF_FLAG}" "${DESTDIR}${COMPFILE#.}" "${COMPFILE}"
+          diff ${DIFF_FLAG} ${DIFF_OPTIONS} "${DESTDIR}${COMPFILE#.}" "${COMPFILE}"
         ) | ${PAGER}
         echo ''
       fi
     else
       echo ''
       echo "  *** There is no installed version of ${COMPFILE}"
+      echo ''
       case "${AUTO_INSTALL}" in
       [Yy][Ee][Ss])
         echo ''
@@ -211,6 +214,12 @@ diff_loop () {
   esac
 }
 
+press_to_continue () {
+  local DISCARD
+  echo -n ' *** Press the [Enter] or [Return] key to continue '
+  read DISCARD
+}
+
 # Set the default path for the temporary root environment
 #
 TEMPROOT='/var/tmp/temproot'
@@ -229,10 +238,11 @@ fi
 
 # Check the command line options
 #
-while getopts ":ascrvhim:t:du:w:D:" COMMAND_LINE_ARGUMENT ; do
+while getopts ":ascrvhipCPm:t:du:w:D:" COMMAND_LINE_ARGUMENT ; do
   case "${COMMAND_LINE_ARGUMENT}" in
   s)
     STRICT=yes
+    unset DIFF_OPTIONS
     ;;
   c)
     DIFF_FLAG='-c'
@@ -257,6 +267,17 @@ while getopts ":ascrvhim:t:du:w:D:" COMMAND_LINE_ARGUMENT ; do
   i)
     AUTO_INSTALL=yes
     ;;
+  C)
+    COMP_CONFS=yes
+    ;;
+  P)
+    PRESERVE_FILES=yes
+    ;;
+  p)
+    PRE_WORLD=yes
+    unset COMP_CONFS
+    unset AUTO_RUN
+    ;;
   m)
     SOURCEDIR=${OPTARG}
     ;;
@@ -282,6 +303,11 @@ while getopts ":ascrvhim:t:du:w:D:" COMMAND_LINE_ARGUMENT ; do
   esac
 done
 
+# Don't force the user to set this in the mergemaster rc file
+if [ -n "${PRESERVE_FILES}" -a -z "${PRESERVE_FILES_DIR}" ]; then
+  PRESERVE_FILES_DIR=/var/tmp/mergemaster/preserved-files-`date +%y%m%d-%H%M%S`
+fi
+
 echo ''
 
 # If the user has a pager defined, make sure we can run it
@@ -352,10 +378,40 @@ DIFF_FLAG=${DIFF_FLAG:--u}
 #
 SOURCEDIR=${SOURCEDIR:-/usr/src/etc}
 
+# Check the width of the user's terminal
+#
+if [ -t 0 ]; then
+  w=`tput columns`
+  case "${w}" in
+  0|'') ;; # No-op, since the input is not valid
+  *)
+    case "${SCREEN_WIDTH}" in
+    '') SCREEN_WIDTH="${w}" ;;
+    "${w}") ;; # No-op, since they are the same
+    *)
+      echo -n "*** You entered ${SCREEN_WIDTH} as your screen width, but stty "
+      echo "thinks it is ${w}."
+      echo ''
+      echo -n "What would you like to use? [${w}] "
+      read SCREEN_WIDTH
+      case "${SCREEN_WIDTH}" in
+      '') SCREEN_WIDTH="${w}" ;;
+      esac
+      ;;
+    esac
+  esac
+fi
+
 # Define what CVS $Id tag to look for to aid portability.
 #
 CVS_ID_TAG=FreeBSD
 
+delete_temproot () {
+  rm -rf "${TEMPROOT}" 2>/dev/null
+  chflags -R 0 "${TEMPROOT}" 2>/dev/null
+  rm -rf "${TEMPROOT}" || exit 1
+}
+
 case "${RERUN}" in
 '')
   # Set up the loop to test for the existence of the
@@ -384,7 +440,7 @@ case "${RERUN}" in
           echo ''
           echo "   *** Deleting the old ${TEMPROOT}"
           echo ''
-          rm -rf "${TEMPROOT}"
+          delete_temproot || exit 1
           unset TEST_TEMP_ROOT
           ;;
         [tT])
@@ -437,28 +493,40 @@ case "${RERUN}" in
   case "${VERBOSE}" in
   '') ;;
   *)
-    echo " *** Press [Enter] or [Return] key to continue"
-    read ANY_KEY
-    unset ANY_KEY
+    press_to_continue
     ;;
   esac
 
-  { cd ${SOURCEDIR} &&
-    case "${DESTDIR}" in
-    '') ;;
-    *)
+  case "${PRE_WORLD}" in
+  '')
+    { cd ${SOURCEDIR} &&
+      case "${DESTDIR}" in
+      '') ;;
+      *)
       make DESTDIR=${DESTDIR} distrib-dirs
-      ;;
-    esac
-    make DESTDIR=${TEMPROOT} distrib-dirs &&
-    make MAKEOBJDIRPREFIX=${TEMPROOT}/usr/obj obj &&
-    make MAKEOBJDIRPREFIX=${TEMPROOT}/usr/obj DESTDIR=${TEMPROOT} \
-        -DNO_MAKEDEV_RUN distribution;} ||
-  { echo '';
-    echo "  *** FATAL ERROR: Cannot 'cd' to ${SOURCEDIR} and install files to";
-    echo "      the temproot environment";
-    echo '';
-    exit 1;}
+        ;;
+      esac
+      make DESTDIR=${TEMPROOT} distrib-dirs &&
+      make MAKEOBJDIRPREFIX=${TEMPROOT}/usr/obj obj &&
+      make MAKEOBJDIRPREFIX=${TEMPROOT}/usr/obj DESTDIR=${TEMPROOT} \
+          distribution;} ||
+    { echo '';
+     echo "  *** FATAL ERROR: Cannot 'cd' to ${SOURCEDIR} and install files to";
+      echo "      the temproot environment";
+      echo '';
+      exit 1;}
+    ;;
+  *)
+    # Only set up files that are crucial to {build|install}world
+    { mkdir -p ${TEMPROOT}/etc &&
+      cp -p ${SOURCEDIR}/master.passwd ${TEMPROOT}/etc &&
+      cp -p ${SOURCEDIR}/group ${TEMPROOT}/etc;} ||
+    { echo '';
+      echo '  *** FATAL ERROR: Cannot copy files to the temproot environment';
+      echo '';
+      exit 1;}
+    ;;
+  esac
 
   # Doing the inventory and removing files that we don't want to compare only
   # makes sense if we are not doing a rerun, since we have no way of knowing
@@ -473,39 +541,36 @@ case "${RERUN}" in
     echo '     However because these files are not updated by this process you'
     echo '     might want to verify their status before rebooting your system.'
     echo ''
-    echo ' *** Press [Enter] or [Return] key to continue'
-    read ANY_KEY
-    unset ANY_KEY
-    diff -qr ${DESTDIR}/etc ${TEMPROOT}/etc | grep "^Only in /etc" | ${PAGER}
+    press_to_continue
+    diff -qr ${DESTDIR}/etc ${TEMPROOT}/etc | grep "^Only in ${DESTDIR}/etc" | ${PAGER}
     echo ''
-    echo ' *** Press [Enter] or [Return] key to continue'
-    read ANY_KEY
-    unset ANY_KEY
+    press_to_continue
     ;;
   esac
 
   # Avoid comparing the motd if the user specifies it in .mergemasterrc
   case "${IGNORE_MOTD}" in
   '') ;;
-  *) rm ${TEMPROOT}/etc/motd
+  *) rm -f ${TEMPROOT}/etc/motd
      ;;
   esac
 
   # Avoid trying to update MAKEDEV if /dev is on a devfs
   if /sbin/sysctl vfs.devfs.generation > /dev/null 2>&1 ; then
-    rm ${TEMPROOT}/dev/MAKEDEV ${TEMPROOT}/dev/MAKEDEV.local
+    rm -f ${TEMPROOT}/dev/MAKEDEV ${TEMPROOT}/dev/MAKEDEV.local
   fi
 
   ;; # End of the "RERUN" test
 esac
 
-# We really don't want to have to deal with these files, since
-# master.passwd is the real file that should be compared, then
-# the user should run pwd_mkdb if necessary.
+# We really don't want to have to deal with files like login.conf.db, pwd.db,
+# or spwd.db.  Instead, we want to compare the text versions, and run *_mkdb.
+# Prompt the user to do so below, as needed.
 #
-[ -f "${TEMPROOT}/etc/spwd.db" ] && rm "${TEMPROOT}/etc/spwd.db"
-[ -f "${TEMPROOT}/etc/passwd" ]  && rm "${TEMPROOT}/etc/passwd"
-[ -f "${TEMPROOT}/etc/pwd.db" ]  && rm "${TEMPROOT}/etc/pwd.db"
+rm -f ${TEMPROOT}/etc/*.db ${TEMPROOT}/etc/passwd
+
+# We only need to compare things like freebsd.cf once
+find ${TEMPROOT}/usr/obj -type f -delete 2>/dev/null
 
 # Get ready to start comparing files
 
@@ -523,6 +588,7 @@ if [ -z "${NEW_UMASK}" -a -z "${AUTO_RUN}" ]; then
     echo "     they are created with by ${SOURCEDIR}/Makefile, compared to"
     echo "     a umask of 022.  This umask allows world read permission when"
     echo "     the file's default permissions have it."
+    echo ''
     echo "     No world permissions can sometimes cause problems.  A umask of"
     echo "     022 will restore the default behavior, but is not mandatory."
     echo "     /etc/master.passwd is a special case.  Its file permissions"
@@ -539,48 +605,73 @@ fi
 
 CONFIRMED_UMASK=${NEW_UMASK:-0022}
 
-# Warn users who still have ${DESTDIR}/etc/sysconfig
 #
-if [ -e "${DESTDIR}/etc/sysconfig" ]; then
+# Warn users who still have old rc files
+#
+for file in atm devfs diskless1 diskless2 isdn network network6 pccard \
+  serial syscons sysctl alpha amd64 i386 ia64 sparc64; do
+  if [ -f "${DESTDIR}/etc/rc.${file}" ]; then
+    OLD_RC_PRESENT=1
+    break
+  fi
+done
+
+case "${OLD_RC_PRESENT}" in
+1)
   echo ''
-  echo " *** There is a sysconfig file on this system in ${DESTDIR}/etc/."
+  echo " *** There are elements of the old rc system in ${DESTDIR}/etc/."
   echo ''
-  echo '     Starting with FreeBSD version 2.2.2 those settings moved from'
-  echo '     /etc/sysconfig to /etc/rc.conf.  If you are upgrading an older'
-  echo '     system make sure that you transfer your settings by hand from'
-  echo '     sysconfig to rc.conf and install the rc.conf file.  If you'
-  echo '     have already made this transition, you should consider'
-  echo '     renaming or deleting the sysconfig file.'
+  echo '     While these scripts will not hurt anything, they are not'
+  echo '     functional on an up to date system, and can be removed.'
   echo ''
+
   case "${AUTO_RUN}" in
   '')
-    echo -n "Continue with the merge process? [yes] "
-    read CONT_OR_NOT
+    echo -n 'Move these files to /var/tmp/mergemaster/old_rc? [yes] '
+    read MOVE_OLD_RC
 
-    case "${CONT_OR_NOT}" in
-    [nN]*)
-      exit 0
-      ;;
+    case "${MOVE_OLD_RC}" in
+    [nN]*) ;;
     *)
-      echo "   *** Continuing"
-      echo ''
+      mkdir -p /var/tmp/mergemaster/old_rc
+        for file in atm devfs diskless1 diskless2 isdn network network6 pccard \
+          serial syscons sysctl alpha amd64 i386 ia64 sparc64; do
+          if [ -f "${DESTDIR}/etc/rc.${file}" ]; then
+            mv ${DESTDIR}/etc/rc.${file} /var/tmp/mergemaster/old_rc/
+          fi
+        done
+      echo '  The files have been moved'
+      press_to_continue
       ;;
     esac
     ;;
   *) ;;
   esac
-fi
+esac
 
 # Use the umask/mode information to install the files
 # Create directories as needed
 #
 do_install_and_rm () {
-  install -m "${1}" "${2}" "${3}" &&
-    if [ -f "${2}" ]; then
-      rm "${2}"
-    else
-      return 0
+  case "${PRESERVE_FILES}" in
+  [Yy][Ee][Ss])
+    if [ -f "${3}/${2##*/}" ]; then
+      mkdir -p ${PRESERVE_FILES_DIR}/${2%/*}
+      cp ${3}/${2##*/} ${PRESERVE_FILES_DIR}/${2%/*}
     fi
+    ;;
+  esac
+
+  install -m "${1}" "${2}" "${3}" &&
+  rm -f "${2}"
+}
+
+# 4095 = "obase=10;ibase=8;07777" | bc
+find_mode () {
+  local OCTAL
+  OCTAL=$(( ~$(echo "obase=10; ibase=8; ${CONFIRMED_UMASK}" | bc) & 4095 &
+    $(echo "obase=10; ibase=8; $(stat -f "%OMp%OLp" ${1})" | bc) )) 
+  printf "%04o\n" ${OCTAL}
 }
 
 mm_install () {
@@ -595,13 +686,11 @@ mm_install () {
   esac
 
   if [ -n "${DESTDIR}${INSTALL_DIR}" -a ! -d "${DESTDIR}${INSTALL_DIR}" ]; then
-    DIR_MODE=`perl -e 'printf "%04o\n", (((stat("$ARGV[0]"))[2] & 07777) &~ \
-      oct("$ARGV[1]"))' "${TEMPROOT}/${INSTALL_DIR}" "${CONFIRMED_UMASK}"`
+    DIR_MODE=`find_mode "${TEMPROOT}/${INSTALL_DIR}"`
     install -d -o root -g wheel -m "${DIR_MODE}" "${DESTDIR}${INSTALL_DIR}"
   fi
 
-  FILE_MODE=`perl -e 'printf "%04o\n", (((stat("$ARGV[0]"))[2] & 07777) &~ \
-      oct("$ARGV[1]"))' "${1}" "${CONFIRMED_UMASK}"`
+  FILE_MODE=`find_mode "${1}"`
 
   if [ ! -x "${1}" ]; then
     case "${1#.}" in
@@ -753,7 +842,8 @@ for COMPFILE in `find . -type f -size +0`; do
     # Do an absolute diff first to see if the files are actually different.
     # If they're not different, delete the one in temproot.
     #
-    if diff -q "${DESTDIR}${COMPFILE#.}" "${COMPFILE}" > /dev/null 2>&1; then
+    if diff -q ${DIFF_OPTIONS} "${DESTDIR}${COMPFILE#.}" "${COMPFILE}" > \
+      /dev/null 2>&1; then
       echo " *** Temp ${COMPFILE} and installed are the same, deleting"
       rm "${COMPFILE}"
     else
@@ -794,7 +884,7 @@ case "${AUTO_RUN}" in
 
   case "${DEL_TEMPROOT}" in
   [yY]*)
-    if rm -rf "${TEMPROOT}"; then
+    if delete_temproot; then
       echo " *** ${TEMPROOT} has been deleted"
     else
       echo " *** Unable to delete ${TEMPROOT}"
@@ -922,5 +1012,52 @@ if [ -r "${MM_EXIT_SCRIPT}" ]; then
   . "${MM_EXIT_SCRIPT}"
 fi
 
+case "${COMP_CONFS}" in
+'') ;;
+*)
+  . ${DESTDIR}/etc/defaults/rc.conf
+
+  (echo ''
+  echo "*** Comparing conf files: ${rc_conf_files}"
+
+  for CONF_FILE in ${rc_conf_files}; do
+    if [ -r "${DESTDIR}${CONF_FILE}" ]; then
+      echo ''
+      echo "*** From ${DESTDIR}${CONF_FILE}"
+      echo "*** From ${DESTDIR}/etc/defaults/rc.conf"
+
+      for RC_CONF_VAR in `grep -i ^[a-z] ${DESTDIR}${CONF_FILE} |
+        cut -d '=' -f 1`; do
+        echo ''
+        grep -w ^${RC_CONF_VAR} ${DESTDIR}${CONF_FILE}
+        grep -w ^${RC_CONF_VAR} ${DESTDIR}/etc/defaults/rc.conf ||
+          echo ' * No default variable with this name'
+      done
+    fi
+  done) | ${PAGER}
+  echo ''
+  ;;
+esac
+
+case "${PRE_WORLD}" in
+'') ;;
+*)
+  MAKE_CONF="${SOURCEDIR%etc}share/examples/etc/make.conf"
+
+  (echo ''
+  echo '*** Comparing make variables'
+  echo ''
+  echo "*** From ${DESTDIR}/etc/make.conf"
+  echo "*** From ${MAKE_CONF}"
+
+  for MAKE_VAR in `grep -i ^[a-z] ${DESTDIR}/etc/make.conf | cut -d '=' -f 1`; do
+    echo ''
+    grep -w ^${MAKE_VAR} ${DESTDIR}/etc/make.conf
+    grep -w ^#${MAKE_VAR} ${MAKE_CONF} ||
+      echo ' * No example variable with this name'
+  done) | ${PAGER}
+  ;;
+esac
+
 exit 0
 
